//
/*! libresignin
*/
// Copyright 2022, Jesús Pérez Lorenzo
//

//! Run with
//!
//! ```not_rust
//! cargo run -p example-static-file-server
//! ```
//! https://rustrepo.com/repo/TrueLayer-reqwest-middleware
//! https://github.com/AscendingCreations/AxumCSRF
//! 

use std::{
	sync::Arc,
	time::Duration,
};
use tokio::sync::Mutex;

use axum::{ 
	error_handling::HandleErrorLayer,
	http::{StatusCode, header::HeaderValue, Method},
	routing::get_service,
	AddExtensionLayer,
  Router,
};
use axum_server::tls_rustls::RustlsConfig;
use tower_http::{
	services::ServeDir,
	trace::TraceLayer,
  cors::{CorsLayer, Origin},
};
use tower::{ServiceBuilder};

use app_env::{
  AppStore,
  appenv::AppEnv,
  appinfo::AppInfo,
  appdata::AppData,
  config::{Config}
};
use connectors::defs::{AppDataConn}; 
use app_auth::AuthStore;
use anyhow::{Result};

//use crate::defs::{DataDBs,CollsData};
use crate::defs::{DataDBs};

// static WEBSERVER: AtomicUsize = AtomicUsize::new(0);
const PKG_VERSION: &'static str = env!("CARGO_PKG_VERSION");
const APP_NAME: &'static str = "libresignin";
// const PKG_VERSION: Option<&'static str> = option_env!("CARGO_PKG_VERSION");
const PKG_NAME: &'static str = env!("CARGO_PKG_NAME");
const PKG_AUTHORS: &'static str = env!("CARGO_PKG_AUTHORS");
const PKG_DESCRIPTION: &'static str = env!("CARGO_PKG_DESCRIPTION");

const AXUM_SESSION_COOKIE_NAME: &str = "axum_session";

pub mod defs;
pub mod kratos;
pub mod handlers;
pub mod utils;
pub mod state;

pub type BxDynResult<T> = std::result::Result<T, Box<dyn std::error::Error>>;

async fn create_auth_store(app_env: &AppEnv,verbose: isize) -> AuthStore {
	let config = app_env.get_curr_websrvr_config();
	let model_path = config.st_auth_model_path();
  let policy_path = config.st_auth_policy_path();
  AuthStore::new(&config,AuthStore::create_enforcer(model_path,policy_path).await,verbose)
}
async fn up_web_server(webpos: usize) -> Result<()> { // WebSettings> {
	let debug = envmnt::get_isize("DEBUG",0);
	let verbose = envmnt::get_isize("WEB_SERVER_VERBOSE", 0);
	let mut app_env = AppEnv::default();
	app_env.curr_web=webpos;
	if verbose > 0 {
		println!("Web services: init {} ___________ ", chrono::Utc::now().timestamp());
	}
	app_env.info = AppInfo::new(
		APP_NAME,
		format!("web: {}",&webpos),
		format!("version: {}",PKG_VERSION),
		format!("authors: {}",PKG_AUTHORS),
		format!("{}",PKG_DESCRIPTION),
	);
	webenv::init_app(&mut app_env,verbose).await.unwrap_or_else(|e| 
		panic!("Error loadding app environment {}",e)
	);
	let config = app_env.get_curr_websrvr_config();
	// let webserver_status = WEBSERVER.load(Ordering::Relaxed);
	let zterton_env = envmnt::get_or(format!("_{}",&config.name).as_str(), "UNKNOWN");
	// if webserver_status != 0 {
	if zterton_env != "UNKNOWN" {
		if verbose  > 0 {
			println!("{} web services at {}",APP_NAME,&zterton_env);
		}
		return Ok(()); 
	}
	// WEBSERVER.store(1,Ordering::Relaxed);
  // TODO pass root file-name frmt from AppEnv Config 
	if verbose > 0 {
		println!("Loading webserver: {} ({})",&config.name,&app_env.curr_web);
	}	
  let (app, socket) = webenv::start_web(&mut app_env).await;
	if verbose > 0 {
 	 println!("Load app store ...");
	}
  // `MemoryStore` just used as an example. Don't use this in production.
  let mem_store = async_session::MemoryStore::new();

  let app_store = AppStore::new(AppData::new(app_env.to_owned(),verbose));
  // As static casbin
	if verbose > 0 {
  	println!("Load auth store ...");
	}
	let auth_store = create_auth_store(&app_env,verbose).await;
	if verbose > 0 {
  	println!("Load data store ...");
	}
	let app_data_conn = AppDataConn::new(APP_NAME.to_string(),app_env.config.datastores_settings.to_owned(),"").await;
	let cookie_store = {
    let file = std::fs::File::open(&config.signin.cookies_path)
      .map(std::io::BufReader::new)
			.unwrap_or_else(|e| 
		    panic!("Error creating cookie path {}: {}",config.signin.cookies_path,e)
			);
			cookie_store::CookieStore::load_json(file)
			.unwrap_or_else(|e| 
		    panic!("Error unable to load path {}: {}",config.signin.cookies_path,e)
			)
  };
  let cookie_store = reqwest_cookie_store::CookieStoreMutex::new(cookie_store); 
	let cookie_store = std::sync::Arc::new(cookie_store);
	let req_cli = reqwest::Client::builder()
//   .timeout(std::time::Duration::from_millis(500))
//	  .connect_timeout(std::time::Duration::from_millis(100))
		.cookie_store(true)
    .build().unwrap_or_else(|e| 
		  panic!("Error creating reqwest client: {}",e)
	);
  let data_dbs = DataDBs {
    // colls: CollsData::new(app_env.to_owned(),verbose),
    app: app_store.to_owned(),
    auth: auth_store.to_owned(),
		conns: Arc::new(Mutex::from(app_data_conn)),
  };
	if verbose > 0 {
 	 println!("Load web filters ...");
	}

  // let us get some static boxes from config values: 
  let log_name = app_env.config.st_log_name();
  // Path for static files 
  let html_path: &str;
	if config.html_path.is_empty() {
    html_path = "/";
	} else {
    html_path = config.st_html_path();
	}
  let html_url: String;
	if config.html_url.is_empty() {
    html_url = String::from("/");
	} else {
    html_url = config.html_url.to_owned();
	}
  // // If not graphQL comment/remove next line 
  // let gql_path = config.st_gql_req_path();
  // // If not graphiQL comment/remove next line Interface GiQL
  // let giql_path = config.st_giql_req_path();

  // let origins: < http::HeaderValue> = config.allow_origin.iter().map(AsRef::as_ref).collect();
  let mut origins: Vec<HeaderValue> = Vec::new();
	for itm in config.allow_origin.to_owned() {
		match HeaderValue::from_str(itm.as_str()) {
     Ok(val) => origins.push(val),
		 Err(e) =>  println!("error {} with {} header for allow_origin",e,itm),
		}
	}	
/*
  let cors = warp::cors()
    //.allow_any_origin()
    .allow_origins(origins)
    //.allow_origins(vec![app_env.config.allow_origin.as_str(), "https://localhost:8000"])
    .allow_credentials(true)
    .allow_header("content-type")
    .allow_header("Authorization")
    .allow_methods(&[Method::GET, Method::POST, Method::DELETE]);

  let auth_api = 
  // Auth routes for login & logout  REQUIRED
  app_auth_filters::auth(app_store.clone(),auth_store.clone(),cors.clone()); // .with(cors.clone());
  
  let gqli_api = 
  // If not graphiQL comment/remove next line Interface GiQL MUST BEFORE graphql post with schema
  // app_api.to_owned()
  graphql::graphiql(gql_path, giql_path, data_dbs.clone()).await;
  if giql_path.len() > 0 && verbose > 0 {
    println!(
     "GraphiQL url: {}://{}:{}/{}",
     &app.protocol, &app.host, &app.port, &giql_path
    );
  } 
	let mut cloud = Cloud::default();
	env_cloud("*", &mut cloud.env).await?;
	load_cloud_env(&mut cloud).await;	
  // If not graphQL comment/remove next line 
  let gql_api=graphql::graphql(gql_path, data_dbs.clone(),cors.clone()).await; //.with(cors.clone());
  // // Add ALL ENTITIES to work with here
  let kloud_api = filters::CollFilters::new(&config.prefix)
		.filters_config(data_dbs.clone(),cloud.clone(),cors.clone());

  let file_api = app_file_filters::files(app_store.clone(),auth_store.clone()).with(cors.clone());
  // Path for static files, better to be LAST
  let fs_api = warp::fs::dir(html_path).with(warp::compression::gzip());

	let home_api = filters::CollFilters::new(&config.prefix)
	  .filters_home(data_dbs.clone(),cloud.clone(),cors.clone(),"info");

  let app_api = auth_api
			.or(gqli_api).or(gql_api)
			.or(home_api)
			.or(kloud_api)
			.or(file_api)
			.or(fs_api)
			.recover(move | error: warp::Rejection| handle_rejection(error, app_store.clone()))
			.boxed();
  // Wrap routes with log to get info
  let routes =  app_api.with(warp::log(log_name));
  //let routes = app_api.with(cors).with(warp::log(log_name));
*/
  println!(
    "Starting http server: {}://{}:{}",
    &app.protocol, &app.host, &app.port
  );
	envmnt::set(APP_NAME, format!("{}:{}",&app.host,&app.port));
	if debug > 0 {
 	  println!("Web services: done {} __________ ",chrono::Utc::now().timestamp());
	}
	let mut web_router = Router::new();
	web_router = crate::handlers::router::router_authz_handlers(web_router,config.signin.authz_routes);
	web_router = crate::handlers::router::router_handlers(web_router);
	if !config.html_path.is_empty() {
		web_router = web_router.nest(
				html_url.as_str(),
				get_service(ServeDir::new(html_path)).handle_error(|error: std::io::Error| async move {
						(
								StatusCode::INTERNAL_SERVER_ERROR,
								format!("Unhandled internal error: {}", error),
						)
				}),
		);
	}
	if config.allow_origin.len() > 0 {
	  web_router = web_router.layer(CorsLayer::new()
		  .allow_origin(Origin::list(origins))
			.allow_methods(vec![Method::GET, Method::POST])
		);
	}
  web_router = web_router.layer(AddExtensionLayer::new(mem_store));
	web_router = web_router.layer(AddExtensionLayer::new(data_dbs));
	web_router = web_router.layer(AddExtensionLayer::new(cookie_store));
	web_router = web_router.layer(TraceLayer::new_for_http());
	web_router = web_router.layer(tower_cookies::CookieManagerLayer::new());
	web_router = web_router.layer(
		 ServiceBuilder::new()
				// Handle errors from middleware
				.layer(HandleErrorLayer::new(handlers::router::handle_error))
				.load_shed()
				.concurrency_limit(1024)
				.timeout(Duration::from_secs(10))
				.layer(TraceLayer::new_for_http())
				.layer(AddExtensionLayer::new(crate::state::SharedState::default()))
				.into_inner(),
	);
	if app.protocol.clone().as_str() == "http" {
		let _ = axum::Server::bind(&socket)
			.serve(web_router.into_make_service())
			.await
			.expect("server failed");
    // let addr = SocketAddr::from(([127, 0, 0, 1], 3000));
    // tracing_subscriber::fmt::init();
    // tracing::debug!("listening on {}", addr);
    //let _ = axum::Server::bind(&addr)
    // let _ = axum::Server::bind(&socket)
    //  .serve(app.into_make_service())
    //   .await;
    //  .unwrap();
	} else {
		match RustlsConfig::from_pem_file(
		  format!("{}/ssl/{}", &config.resources_path, "fullchain.pem")
		  ,format!("{}/ssl/{}", &config.resources_path,"privkey.pem")
		).await {
      Ok(ssl_cfg) => {
				let _ = axum_server::bind_rustls(socket,ssl_cfg)
					.serve(web_router.into_make_service())
					.await
					.expect("server failed");
			},
			Err(e) => {
				println!("SSL Certifcates error: {}",e);
				return Ok(());
			}
		};
	}
	Ok(())
}
// fn get_args() -> (String,String) {
// 	let args: Vec<String> = std::env::args().collect();
// 	let mut arg_cfg_path = String::from("");
// 	let mut arg_env_path = String::from("");
// 	args.iter().enumerate().for_each(|(idx,arg)| {
// 		if arg == "-c" { 
// 			arg_cfg_path=args[idx+1].to_owned();
// 		} else if arg == "-e" { 
// 			arg_env_path=args[idx+1].to_owned();
// 		}
//   });
// 	(arg_cfg_path,arg_env_path)
// }
//async fn get_app_env(arg_cfg_path: String,verbose: isize) -> Result<(Cloud,AppEnv)> {
//	let mut cloud = Cloud::default();
//	load_cloud_env(&mut cloud).await;	
// async fn get_app_env(arg_cfg_path: String,verbose: isize) -> Result<AppEnv> {
// 	let mut app_env = AppEnv::default();
// 	let config_content = Config::load_file_content(verbose,&arg_cfg_path);
//   if ! config_content.contains("run_mode") {
// 		Err(anyhow!("Run mode not found in config {}", &arg_cfg_path))
// 	} else {
// 		app_env.config = Config::new(config_content,verbose);
// 		// Ok((cloud,app_env))
// 		Ok(app_env)
// 	}
// }
/*
async fn set_reqenv(app_env: &AppEnv,verbose: isize) -> ReqEnv {
	let app_store =	AppStore::new(AppData::new(app_env.to_owned(),verbose));
	let auth_store = create_auth_store(&app_env,verbose).await;
	let mut headers = HeaderMap::new();
  headers.insert(http::header::HOST, "localhost".parse().unwrap());
	ReqEnv::new( 
		app_store, 
		auth_store,
		headers,
	  Method::GET,
		"/config", "config", "kloud"
	)
}
*/
#[tokio::main]
async fn main() -> BxDynResult<()> {  //std::io::Result<()> { 
	let args: Vec<String> = std::env::args().collect();
	if args.len() > 1 {
		match args[1].as_str() {
			"-h" | "--help" => {
	  		println!("{} USAGE: -c config-toml -e env.file",PKG_NAME);
				return Ok(());
			},
			"-v" | "--version" => {
	  		println!("{} version: {}",PKG_NAME,PKG_VERSION);
				return Ok(());
			},
			_ => println!("{}",PKG_NAME),
		}
	}
	let debug=envmnt::get_isize("DEBUG",0);
	let mut arg_cfg_path = String::from("");
	let mut arg_env_path = String::from("");
	args.iter().enumerate().for_each(|(idx,arg)| {
		if arg == "-c" { 
			arg_cfg_path=args[idx+1].to_owned();
		} else if arg == "-e" { 
			arg_env_path=args[idx+1].to_owned();
		}
  });
	if !arg_env_path.is_empty() {
		let env_path = std::path::Path::new(&arg_env_path);
		dotenv::from_path(env_path)?;
	}
	pretty_env_logger::init();
	let config_content = Config::load_file_content(debug, &arg_cfg_path);
	if !config_content.contains("run_mode") {
		panic!("Error no run_mode found or config path incomplete");
	}
	let config = Config::new(config_content,debug);
	let app_data_conn = AppDataConn::new(APP_NAME.to_string(),config.datastores_settings.to_owned(),"").await;
	if config.datastores_settings.len() > 0 {
		if !app_data_conn.check_connections(config.datastores_settings.to_owned()).await {
			println!("Error checking app data store connections");
		} 
  }
	// Set the RUST_LOG, if it hasn't been explicitly defined
	if std::env::var_os("RUST_LOG").is_none() {
			std::env::set_var(
					"RUST_LOG",
					"example_static_file_server=debug,tower_http=debug",
			)
	}
	if config.run_websrvrs {
		for (pos,it) in config.websrvrs.iter().enumerate() {
			if debug > 1 {
				println!("{} -> {}",it.name,pos);
			}
			//tokio::join!(async move {
			let handle = tokio::spawn(async move { up_web_server(pos).await });
			let _ = handle.await;
			tokio::time::sleep(tokio::time::Duration::from_secs(3)).await;
		}
	}
   Ok(())
}
